"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ViewportList = void 0;
var jsx_runtime_1 = require("react/jsx-runtime");
var react_1 = require("react");
var IS_SSR = typeof window === 'undefined';
var IS_TOUCH_DEVICE = IS_SSR
    ? false
    : (function () {
        try {
            return 'ontouchstart' in window || navigator.maxTouchPoints;
        }
        catch (error) {
            return false;
        }
    })();
var IS_OVERFLOW_ANCHOR_SUPPORTED = IS_SSR
    ? false
    : (function () {
        try {
            return window.CSS.supports('overflow-anchor: auto');
        }
        catch (error) {
            return false;
        }
    })();
var PROP_NAME_FOR_Y_AXIS = {
    top: 'top',
    bottom: 'bottom',
    clientHeight: 'clientHeight',
    scrollHeight: 'scrollHeight',
    scrollTop: 'scrollTop',
    overflowY: 'overflowY',
    height: 'height',
    minHeight: 'minHeight',
    maxHeight: 'maxHeight',
    marginTop: 'marginTop',
};
var PROP_NAME_FOR_X_AXIS = {
    top: 'left',
    bottom: 'right',
    scrollHeight: 'scrollWidth',
    clientHeight: 'clientWidth',
    scrollTop: 'scrollLeft',
    overflowY: 'overflowX',
    minHeight: 'minWidth',
    height: 'width',
    maxHeight: 'maxWidth',
    marginTop: 'marginLeft',
};
var getStyle = function (propName, size, marginTop) {
    var _a;
    if (marginTop === void 0) { marginTop = 0; }
    return (_a = {
            padding: 0,
            margin: 0,
            border: 'none',
            visibility: 'hidden',
            // We need to off 'overflow-anchor' for spacers. Otherwise, it may cause unnecessary scroll jumps.
            overflowAnchor: 'none'
        },
        _a[propName.minHeight] = size,
        _a[propName.height] = size,
        _a[propName.maxHeight] = size,
        _a[propName.marginTop] = marginTop,
        _a);
};
var normalizeValue = function (min, value, max) {
    if (max === void 0) { max = Infinity; }
    return Math.max(Math.min(value, max), min);
};
var useIsomorphicLayoutEffect = IS_SSR ? react_1.useEffect : react_1.useLayoutEffect;
var ViewportListInner = function (_a, ref) {
    var _b = _a.viewportRef, viewportRef = _b === void 0 ? { current: IS_SSR ? null : document.documentElement } : _b, _c = _a.items, items = _c === void 0 ? [] : _c, _d = _a.itemMinSize, itemMinSize = _d === void 0 ? 0 : _d, _e = _a.margin, margin = _e === void 0 ? -1 : _e, _f = _a.overscan, overscan = _f === void 0 ? 1 : _f, _g = _a.axis, axis = _g === void 0 ? 'y' : _g, _h = _a.initialIndex, initialIndex = _h === void 0 ? -1 : _h, _j = _a.initialAlignToTop, initialAlignToTop = _j === void 0 ? true : _j, _k = _a.initialOffset, initialOffset = _k === void 0 ? 0 : _k, children = _a.children, onViewportIndexesChange = _a.onViewportIndexesChange, _l = _a.overflowAnchor, overflowAnchor = _l === void 0 ? 'auto' : _l, _m = _a.withCache, withCache = _m === void 0 ? true : _m;
    var propName = axis === 'y' ? PROP_NAME_FOR_Y_AXIS : PROP_NAME_FOR_X_AXIS;
    var maxIndex = items.length - 1;
    var _o = (0, react_1.useState)(function () { return [
        normalizeValue(0, itemMinSize),
        normalizeValue(-1, margin),
    ]; }), _p = _o[0], itemHeight = _p[0], itemMargin = _p[1], setItemDimensions = _o[1];
    var itemHeightWithMargin = normalizeValue(0, itemHeight + itemMargin);
    var overscanSize = normalizeValue(0, Math.ceil(overscan * itemHeightWithMargin));
    var _q = (0, react_1.useState)([initialIndex, initialIndex]), indexes = _q[0], setIndexes = _q[1];
    var startIndex = (indexes[0] = normalizeValue(0, indexes[0], maxIndex));
    var endIndex = (indexes[1] = normalizeValue(startIndex, indexes[1], maxIndex));
    var topSpacerRef = (0, react_1.useRef)(null);
    var bottomSpacerRef = (0, react_1.useRef)(null);
    var cacheRef = (0, react_1.useRef)([]);
    var scrollToIndexRef = (0, react_1.useRef)(initialIndex >= 0 ? { index: initialIndex, alignToTop: initialAlignToTop, offset: initialOffset } : null);
    var marginTopRef = (0, react_1.useRef)(0);
    var viewportIndexesRef = (0, react_1.useRef)([-1, -1]);
    var anchorElementRef = (0, react_1.useRef)(null);
    var minAverageSizeRef = (0, react_1.useRef)(Infinity);
    var stepRef = (0, react_1.useRef)(function () { });
    var topSpacerStyle = (0, react_1.useMemo)(function () {
        return getStyle(propName, 
        // Array.prototype.reduce() runs only for initialized items.
        cacheRef.current
            .slice(0, startIndex)
            .reduce(function (sum, next) { return sum + (next - itemHeight); }, startIndex * itemHeightWithMargin), marginTopRef.current);
    }, [propName, startIndex, itemHeightWithMargin, itemHeight]);
    var bottomSpacerStyle = (0, react_1.useMemo)(function () {
        return getStyle(propName, 
        // Array.prototype.reduce() runs only for initialized items.
        cacheRef.current
            .slice(endIndex + 1, maxIndex + 1)
            .reduce(function (sum, next) { return sum + (next - itemHeight); }, itemHeightWithMargin * (maxIndex - endIndex)));
    }, [propName, endIndex, maxIndex, itemHeightWithMargin, itemHeight]);
    stepRef.current = function () {
        var _a, _b;
        var viewport = viewportRef.current;
        var topSpacer = topSpacerRef.current;
        var bottomSpacer = bottomSpacerRef.current;
        if (!viewport || !topSpacer || !bottomSpacer) {
            return;
        }
        var topElement = topSpacer.nextSibling;
        var bottomElement = bottomSpacer.previousSibling;
        var topSecondElement = topElement === bottomSpacer ? bottomSpacer : topElement.nextSibling;
        var bottomSecondElement = bottomElement === topSpacer ? topSpacer : bottomElement.previousSibling;
        var viewportRect = viewport.getBoundingClientRect();
        var topSpacerRect = topSpacer.getBoundingClientRect();
        var bottomSpacerRect = bottomSpacer.getBoundingClientRect();
        var limits = (_a = {},
            _a[propName.top] = viewport === document.documentElement ? 0 : viewportRect[propName.top],
            _a[propName.bottom] = viewport === document.documentElement
                ? document.documentElement.clientHeight
                : viewportRect[propName.bottom],
            _a);
        var limitsWithOverscanSize = (_b = {},
            _b[propName.top] = limits[propName.top] - overscanSize,
            _b[propName.bottom] = limits[propName.bottom] + overscanSize,
            _b);
        if ((marginTopRef.current < 0 &&
            topSpacerRect[propName.top] - marginTopRef.current >= limitsWithOverscanSize[propName.top]) ||
            (marginTopRef.current > 0 && topSpacerRect[propName.top] >= limitsWithOverscanSize[propName.top])) {
            topSpacer.style[propName.marginTop] = '0px';
            viewport.style[propName.overflowY] = 'hidden';
            viewport[propName.scrollTop] += -marginTopRef.current;
            viewport.style[propName.overflowY] = '';
            marginTopRef.current = 0;
            return;
        }
        if (itemHeight === 0 || itemMargin === -1) {
            if (topElement === bottomSpacer) {
                return;
            }
            var nextItemHeight = itemHeight === 0 ? topElement[propName.clientHeight] : itemHeight;
            var nextItemMargin = itemMargin === -1
                ? bottomSpacerRect[propName.top] - topSpacerRect[propName.bottom] - nextItemHeight
                : itemMargin;
            setItemDimensions([nextItemHeight, nextItemMargin]);
            return;
        }
        minAverageSizeRef.current = Math.min(minAverageSizeRef.current, Math.ceil((bottomSpacerRect[propName.top] - topSpacerRect[propName.bottom]) / (endIndex + 1 - startIndex)) || Infinity);
        var nextStartIndex = startIndex;
        var nextEndIndex = endIndex;
        if (scrollToIndexRef.current) {
            if (marginTopRef.current) {
                topSpacer.style[propName.marginTop] = '0px';
                viewport.style[propName.overflowY] = 'hidden';
                viewport[propName.scrollTop] += -marginTopRef.current;
                viewport.style[propName.overflowY] = '';
                marginTopRef.current = 0;
                return;
            }
            var targetIndex = normalizeValue(0, scrollToIndexRef.current.index, maxIndex);
            if (targetIndex < startIndex || targetIndex > endIndex) {
                setIndexes([targetIndex, targetIndex]);
                return;
            }
            var index = startIndex;
            var element = topElement;
            while (element && element !== bottomSpacer) {
                if (index === targetIndex) {
                    element.scrollIntoView(scrollToIndexRef.current.alignToTop);
                    if (scrollToIndexRef.current.offset) {
                        viewport[propName.scrollTop] += scrollToIndexRef.current.offset;
                    }
                    scrollToIndexRef.current = null;
                    break;
                }
                index++;
                element = element.nextSibling;
            }
            return;
        }
        // If top spacer is intersecting viewport
        if (topSpacerRect[propName.bottom] >= limitsWithOverscanSize[propName.top]) {
            var diff = Math.ceil((topSpacerRect[propName.bottom] - limitsWithOverscanSize[propName.top]) / minAverageSizeRef.current);
            nextStartIndex -= diff;
            // If bottom second element is not intersecting viewport
            if (bottomSecondElement !== topSpacer &&
                bottomSecondElement.getBoundingClientRect()[propName.bottom] > limitsWithOverscanSize[propName.bottom]) {
                var index = endIndex;
                var element = bottomElement;
                while (element && element !== topSpacer) {
                    if (element.getBoundingClientRect()[propName.bottom] <= limitsWithOverscanSize[propName.bottom]) {
                        nextEndIndex = index + 1;
                        break;
                    }
                    index--;
                    element = element.previousSibling;
                }
                if (nextEndIndex === endIndex) {
                    nextEndIndex -= diff;
                }
            }
        }
        // If bottom spacer is intersecting viewport
        if (bottomSpacerRect[propName.top] <= limitsWithOverscanSize[propName.bottom]) {
            var diff = Math.ceil((limitsWithOverscanSize[propName.bottom] - bottomSpacerRect[propName.top]) / minAverageSizeRef.current);
            nextEndIndex += diff;
            // If top second element is not intersecting viewport
            if (topSecondElement !== bottomSpacer &&
                topSecondElement.getBoundingClientRect()[propName.top] < limitsWithOverscanSize[propName.top]) {
                var index = startIndex;
                var element = topElement;
                while (element && element !== bottomSpacer) {
                    if (element.getBoundingClientRect()[propName.top] >= limitsWithOverscanSize[propName.top]) {
                        nextStartIndex = index - 1;
                        break;
                    }
                    index++;
                    element = element.nextSibling;
                }
                if (nextStartIndex === startIndex) {
                    nextStartIndex += diff;
                }
            }
        }
        if (onViewportIndexesChange) {
            var index = startIndex;
            var element = topElement;
            var startViewportIndex = startIndex;
            while (element && element !== bottomSpacer) {
                if (element.getBoundingClientRect()[propName.bottom] > limits[propName.top]) {
                    startViewportIndex = index;
                    break;
                }
                index++;
                element = element.nextSibling;
            }
            index = endIndex;
            element = bottomElement;
            var endViewportIndex = endIndex;
            while (element && element !== topSpacer) {
                if (element.getBoundingClientRect()[propName.top] < limits[propName.bottom]) {
                    endViewportIndex = index;
                    break;
                }
                index--;
                element = element.previousSibling;
            }
            if (startViewportIndex !== viewportIndexesRef.current[0] ||
                endViewportIndex !== viewportIndexesRef.current[1]) {
                viewportIndexesRef.current = [startViewportIndex, endViewportIndex];
                onViewportIndexesChange(viewportIndexesRef.current);
            }
        }
        nextStartIndex = normalizeValue(0, nextStartIndex, maxIndex);
        nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);
        if (nextStartIndex === startIndex && nextEndIndex === endIndex) {
            return;
        }
        if (nextStartIndex !== startIndex) {
            var anchorIndex = Math.max(nextStartIndex, startIndex);
            var anchorElement = null;
            if (anchorIndex === startIndex && anchorIndex <= nextEndIndex) {
                anchorElement = topElement;
            }
            else if (anchorIndex === nextStartIndex && anchorIndex <= endIndex) {
                var index = startIndex;
                var element = topElement;
                while (element && element !== bottomSpacer) {
                    if (index === anchorIndex) {
                        anchorElement = element;
                        break;
                    }
                    if (withCache && element[propName.clientHeight] !== itemHeight) {
                        cacheRef.current[index] = element[propName.clientHeight];
                    }
                    index++;
                    element = element.nextSibling;
                }
            }
            anchorElementRef.current = anchorElement;
        }
        setIndexes([nextStartIndex, nextEndIndex]);
    };
    useIsomorphicLayoutEffect(function () {
        var frameId;
        var frame = function () {
            frameId = requestAnimationFrame(frame);
            stepRef.current();
        };
        frame();
        return function () { return cancelAnimationFrame(frameId); };
    }, []);
    var anchorScrollTopOnRender;
    var anchorHeightOnRender;
    if (anchorElementRef.current && viewportRef.current && topSpacerRef.current) {
        anchorScrollTopOnRender = viewportRef.current[propName.scrollTop];
        anchorHeightOnRender =
            anchorElementRef.current.getBoundingClientRect()[propName.top] -
                topSpacerRef.current.getBoundingClientRect()[propName.top];
    }
    useIsomorphicLayoutEffect(function () {
        var anchorElement = anchorElementRef.current;
        anchorElementRef.current = null;
        var viewport = viewportRef.current;
        var topSpacer = topSpacerRef.current;
        if ((IS_OVERFLOW_ANCHOR_SUPPORTED && overflowAnchor !== 'none') ||
            !anchorElement ||
            !viewport ||
            !topSpacer ||
            anchorScrollTopOnRender === undefined ||
            anchorHeightOnRender === undefined ||
            anchorScrollTopOnRender !== viewport[propName.scrollTop]) {
            return;
        }
        var offset = anchorElement.getBoundingClientRect()[propName.top] -
            topSpacer.getBoundingClientRect()[propName.top] -
            anchorHeightOnRender;
        if (!offset) {
            return;
        }
        if (IS_TOUCH_DEVICE) {
            marginTopRef.current -= offset;
            topSpacer.style[propName.marginTop] = marginTopRef.current + "px";
            return;
        }
        viewport[propName.scrollTop] += offset;
    }, [startIndex]);
    (0, react_1.useImperativeHandle)(ref, function () { return ({
        scrollToIndex: function (index, alignToTop, offset) {
            if (index === void 0) { index = -1; }
            if (alignToTop === void 0) { alignToTop = true; }
            if (offset === void 0) { offset = 0; }
            scrollToIndexRef.current = { index: index, alignToTop: alignToTop, offset: offset };
        },
    }); }, []);
    return ((0, jsx_runtime_1.jsxs)(react_1.Fragment, { children: [(0, jsx_runtime_1.jsx)("div", { ref: topSpacerRef, style: topSpacerStyle }, void 0), items.slice(startIndex, endIndex + 1).map(function (item, index) { return children(item, startIndex + index, items); }), (0, jsx_runtime_1.jsx)("div", { ref: bottomSpacerRef, style: bottomSpacerStyle }, void 0)] }, void 0));
};
exports.ViewportList = (0, react_1.forwardRef)(ViewportListInner);
// eslint-disable-next-line import/no-default-export
exports.default = exports.ViewportList;
